function [yhat_RWI,yhat_RI] = iterVSTpoissonDeb(z,v)
% Iterative Poisson deblurring algorithm based on variance stabilization and colored noise removal (v1.00)
%
% INPUT:   z   Poisson noisy realization of the blurred image g=y*v:  z ~ P(g)
%          v   point-spread function (PSF)
%
% OUTPUT:  yhat_RWI  deblurred image,  estimate of y, after second stage (Wiener regularized inverse)
%          yhat_RI   deblurred image,  estimate of y, after first stage (standard regularized inverse)
%
% L. Azzari and A. Foi, "Variance Stabilization in Poisson Image Deblurring", Proc. IEEE ISBI 2017
%
%  v1.00, April 18, 2017
%
% http://www.cs.tut.fi/~foi/invansc/
%
%  L. Azzari and A. Foi - Tampere University of Technology - 2017 - All rights reserved.
% --------------------------------------------------------------------------------------------------
%
%% Parameter setup
load paramsFromQstdFunDeb                    % provide algorithm parameters based on histogram (quantiles) of observed image and st.dev. of PSF
Qs=linspace(0,1,numel(whichQs));             % quantiles of interest  (typically first and last, i.e. min and max are ignored, as they are unstable)
QQ=quantile(z(:),Qs(whichQs));               % vector of quantiles of z
paramsFromQv=paramConstraints(paramInterp(paramsFromQstdFunDeb(PsDeb(QQ)),getStD(v))); % compute parameters based on quantiles and st.dev. of PSF
lambda_K=paramsFromQv(1);                    % last value of lambda
K=paramsFromQv(2);                           % number of iterations
h_1=1+2*paramsFromQv(3);                     % bin size at iteration 1, i.e. h_1
h_K=1+2*paramsFromQv(4);                     % bin size at last iteration, i.e. h_K
eps_RI  = paramsFromQv(5);                   % regularization parameter for the standard regularized inverse
eps_RWI = paramsFromQv(6);                   % regularization parameter for the Wiener regularized inverse
lambda_1=1;                                  % first value of lambda
delta=127/40;                                % delta for the regularized inverse


%% compute denoising parameters for intermediate iterations
if K>1
    lambdaS=linspace(lambda_1,lambda_K,K);    % lambdas
else
    lambdaS=lambda_1;
    lambda_K=lambda_1;
end
hS=max(h_K,h_1-2*(1:K)+2);     % bin sizes
% remove steps with too small lambda, and update end values
hS(lambdaS<0.01)=[];
lambdaS(lambdaS<0.01)=[];
lambda_K=lambdaS(end);
% h_K=hS(end);
if lambda_K>0.99 % if last lambda is very close to 1, then do only last iteration, because previous ones have negligible effect in the convex combination
    hS=h_K;
    lambdaS=1;
    % lambda_K=1;
end

%% Algorithm
% print out lambda and h parameters to be used during the iterations
disp([['  + lambda:  ';'  | h:       '],num2str([lambdaS;hS])])
disp(['  | eps_RI:  ' num2str(eps_RI,4) '   eps_RWI:  ' num2str(eps_RWI,4)])

%% Anscombe VST
f=@(x,c) 2*sqrt(max(0,x*c+3/8));

%% Fourier transform of the PSF
[Xv, Xh]            = size(z);
[ghy,ghx]           = size(v);
big_v               = zeros(Xv,Xh);
big_v(1:ghy,1:ghx)  = v;
big_v               = circshift(big_v,-round([(ghy-1)/2 (ghx-1)/2])); % pad PSF with zeros to whole image domain, and center it
V                   = fft2(big_v); % frequency response of the PSF


%% two-stage deblurring (RI and RWI)
timeOld=now;
Gamma_delta=delta*abs(V)*sqrt(numel(z)*sum(v(:))/sum(z(:)))-abs(V).^2;
for ind_stage=1:2
    if ind_stage == 1
        %% Step 1 Regularized Inversion
        EPSILON2=max(0,min(eps_RI^2*sum(z(:)),Gamma_delta));
    else
        %% Step 2 Regularized Wiener Inversion
        yhat_RI = yhat0;
        EPSILON2=max(0,min(eps_RWI^2*sum(z(:))*abs(fft2(double(yhat0))).^-2,Gamma_delta));
    end
    RI= conj(V)./( (abs(V).^2) + EPSILON2);
    RI(1)=1;
    zRI=real(ifft2(fft2(z).*RI));   % RI OBSERVATION
    yhat0=zRI;
    for i=1:numel(lambdaS)
        lambda=lambdaS(i);    % lambda for current iteration
        h=hS(i);              % bin size for current iteration
        if lambda>0
            if i>1
                z_i=lambda*zRI+(1-lambda)*yhat0;
            else
                z_i=zRI;
            end
            %% Binning
            z_B=bin_B_h(z_i,h);
            
            %% Compute PSD encompassing RI and binning
            [Xv, Xh] = size(z_i);  Xvp=ceil(Xv/h)*h;  Xhp=ceil(Xh/h)*h;
            big_Bin = zeros(Xvp,Xhp);big_Bin(1:h,1:h)=1;  % binning kernel (uniform with unit coefficients)
            PSDcv=fftshift(abs((fft2(ifftshift(ifft2(RI)),Xvp,Xhp).*fft2(big_Bin))).^2); % convolve with binning kernel and center PSD
            PSDcv=PSDcv/(Xvp*Xhp/numel(RI));
            PSI=0;     % Initiliaze noise power spectral density (PSD)
            % shift-fold PSD and accumulate
            hh=floor(h/2); hH=h-hh-1;
            for j1=-hh:hH  % vertical iterative cyclic shift
                PSDcv=circshift(PSDcv,[Xvp/h 0]);
                PSDcvh=PSDcv;
                for j2=-hh:hH  % horizontal iterative cyclic shift
                    PSDcvh=circshift(PSDcvh,[0 Xhp/h]);
                    ddv=round((Xvp-Xvp/h)/2); ddh=round((Xhp-Xhp/h)/2);
                    PSI=PSI+PSDcvh(ddv+1:ddv+Xvp/h,ddh+1:ddh+Xhp/h);  % accumulate folded central part
                end
            end
            PSI=ifftshift(PSI)*(numel(PSI)/(h^2))/numel(z_i);  % shifts and scales PSD
            c_i=sqrt(PSI(1))/(lambda^2*mean(abs(PSI(:))));  % scaling factor to make expectation and variance equal
            
            %% Apply forward VST
            fz = f(z_B,c_i);
            sigma_den = 1;  % noise standard-deviation assumed after variance stabiliation
            
            %% Scale the image to [0.05 0.95] range (BM3D requires input within [0 1])
            scale_range = 0.9;
            scale_shift = (1-scale_range)/2;
            maxzans = max(fz(:))+eps;
            minzans = min(fz(:))-eps;
            fz   =   (fz-minzans)/(maxzans-minzans);
            sigma_den = sigma_den/(maxzans-minzans);
            fz = scale_shift + fz*scale_range;
            sigma_den = sigma_den*scale_range;
            PSI=PSI * (sigma_den^2 / mean(abs(PSI(:))) * numel(fz));  %normalization of PSI to make the pixelwise noise variance sigma_den^2
            %% Denoising by BM3D assuming signal-independent colored Gaussian noise http://www.cs.tut.fi/~foi/GCF-BM3D/
            D_i = BM3D_COL(fz,PSI,'np');
            
            %% rescaling of the image to its original range
            D_i = (D_i-scale_shift)/scale_range;
            D_i = D_i*(maxzans-minzans)+minzans;
            yhat0 = Anscombe_inverse_exact_unbiased(D_i)/c_i;
            
            %% Debinning
            J=9;
            yhat0=debin_Binv_h(yhat0,size(zRI),h,J);
        end
    end
end
yhat_RWI = yhat0;

timeNow=now;
disp(['  + Elapsed ', num2str(86400*(timeNow-timeOld)),'s']);
